AC_DEFUN([NORDUGRID_INIT],
[

AC_MSG_CHECKING([for SYSV init scripts])
AC_ARG_WITH(sysv-scripts-location,
  [  --sysv-scripts-location=<PATH>       Location of the SYSV init scripts (e.g. /etc/init.d)],
  [ initddir="$withval" ],
  [
    initdir=
    for i in /etc/init.d /etc/rc.d/init.d; do
      if test -d "$i" ; then
	initddir="$i"
	break
      fi
    done
  ]
)
if test ! -d "$initddir"; then
  AC_MSG_ERROR(could not find a suitable location for the SYSV init scripts)
fi
AC_MSG_RESULT($initddir)
AC_SUBST(initddir)

AC_ARG_WITH(monitor-prefix,
  [  --with-monitor-prefix=<PATH>         Specify the location of the Grid Monitor WWW scripts],
  [ monitor_prefix="$withval" ],
  [ monitor_prefix="/var/www/html/gridmonitor" ]
)
AC_SUBST(monitor_prefix)

AC_ARG_WITH(monitor-local-prefix,
  [  --with-monitor-local-prefix=<PATH>   Specify the relative location of the Grid Monitor WWW scripts],
  [ monitor_local_prefix="$withval" ],
  [ monitor_local_prefix=`basename $monitor_prefix` ]
)
AC_SUBST(monitor_local_prefix)

AC_ARG_WITH(gsoap-location,
  [  --with-gsoap-location=<PATH>         Specify the gSOAP installation path],
  [ AC_PATH_PROG(soapcpp2, soapcpp2, not found, $with_gsoap_location/bin) ],
  [ AC_PATH_PROG(soapcpp2, soapcpp2, not found) ]
)
if test "x$soapcpp2" = "xnot found"; then
  gsoap_location=""
else
  gsoap_location=`echo "$soapcpp2" | sed 's%/bin/soapcpp2$%%;t;s%.*%%'`
  SOAPCPP2=$gsoap_location/bin/soapcpp2
  if test "x$gsoap_location" = "x/usr"; then
    :
  else
    GSOAP_INCLUDES=-I$gsoap_location/include
    AC_SUBST(GSOAP_INCLUDES)
  fi
  GSOAP_LIBS=$gsoap_location/lib/libgsoap++.a
  AC_DEFINE(HAVE_GSOAP,,[HAVE_GSOAP])
fi
AC_SUBST(SOAPCPP2)
AC_SUBST(GSOAP_LIBS)

AC_ARG_WITH(mysql-location,
  [  --with-mysql-location=<PATH>         Specify the MySQL installation path],
  [ AC_PATH_PROG(mysql_config, mysql_config, not found, $with_mysql_location/bin) ],
  [ AC_PATH_PROG(mysql_config, mysql_config, not found) ]
)
if test "x$mysql_config" != "xnot found"; then
  AC_MSG_CHECKING(for mysql includes)
  [ mysql_include=`$mysql_config --cflags | sed "s/^-I\(.*\)/\1\/mysql.h/;s/'//g"` ]
  if ! test -f $mysql_include; then
    AC_MSG_RESULT(not found)
    mysql_config="not found"
  else
    AC_MSG_RESULT(found)
    MYSQL_INCLUDES=`$mysql_config --cflags | sed "s/'//g"`
    MYSQL_LIBS=`$mysql_config --libs | sed "s/'//g"`
  fi
fi
AC_SUBST(MYSQL_INCLUDES)
AC_SUBST(MYSQL_LIBS)

AC_ARG_WITH(xml-location,
  [  --with-libxml-location=<PATH>        Specify the libxml installation path],
  [ AC_PATH_PROG(xml_config, xml-config, not found, $with_libxml_location/bin) ],
  [ AC_PATH_PROG(xml_config, xml-config, not found) ]
)
if test "x$xml_config" != "xnot found"; then
  XML_INCLUDES=`$xml_config --cflags`
  XML_LIBS=`$xml_config --libs`
fi
AC_SUBST(XML_INCLUDES)
AC_SUBST(XML_LIBS)
AM_CONDITIONAL(HAVE_LIBXML, test "x$xml_config" != "x")

if test ! "x$xml_config" = "x"; then
  AC_MSG_CHECKING(for libxml includes)
  [ xml_include_dir=`echo $XML_INCLUDES | sed 's/^.*-I\([^ ]*\).*/\1/'` ]
  if ! test -d $xml_include_dir/libxml; then
    AC_MSG_RESULT(old style)
    AC_DEFINE(HAVE_OLD_LIBXML_INCLUDES,,[HAVE_OLD_LIBXML_INCLUDES])
  else
    AC_MSG_RESULT(new style)
  fi
fi

AM_CONDITIONAL(BUILD_LOGGER, test "x$mysql_config" != "x" && test "x$gsoap_location" != "x")
AM_CONDITIONAL(BUILD_HTTPSD, test "x$gsoap_location" != "x")
AM_CONDITIONAL(BUILD_SE, test "x$xml_config" != "x" && test "x$gsoap_location" != "x")

AC_ARG_WITH(nordugrid-location,
  [  --with-nordugrid-location=<PATH>     Specify the nordugrid installation path],
  [ prefix=$with_nordugrid_location ],
  [ prefix=/usr ]
)

AC_MSG_CHECKING(for GLOBUS_LOCATION)
AC_ARG_WITH(globus-location,
  [  --with-globus-location=<PATH>        Specify the globus installation path],
  [
    GLOBUS_LOCATION="$withval"
    export GLOBUS_LOCATION
  ],
  [
    if test "x$GLOBUS_LOCATION" = "x"; then
      GLOBUS_LOCATION=/opt/globus
      export GLOBUS_LOCATION
    fi
  ]
)
if test ! -d $GLOBUS_LOCATION; then
  AC_MSG_ERROR(could not find GLOBUS_LOCATION)
fi
AC_MSG_RESULT($GLOBUS_LOCATION)  
AC_SUBST(GLOBUS_LOCATION)

AC_MSG_CHECKING(for GPT_LOCATION)
AC_ARG_WITH(gpt-location,
  [  --with-gpt-location=<PATH>           Specify the gpt installation path],
  [
    if test "x$withval" = "x"; then
      if test "x$GPT_LOCATION" = "x"; then
	GPT_LOCATION=/opt/gpt
	if test ! -d $GPT_LOCATION; then
	  GPT_LOCATION=$GLOBUS_LOCATION
	fi
	export GPT_LOCATION
      fi
    else
      GPT_LOCATION="$withval"
      export GPT_LOCATION
    fi
  ],
  [
    if test "x$GPT_LOCATION" = "x"; then
      GPT_LOCATION=/opt/gpt
      if test ! -d $GPT_LOCATION; then
	GPT_LOCATION=$GLOBUS_LOCATION
      fi
      export GPT_LOCATION
    fi
  ]
)
AC_MSG_RESULT($GPT_LOCATION)
AC_SUBST(GPT_LOCATION)

AC_ARG_WITH(globus-makefile-header,
  [  --with-globus-makefile-header        Use globus-makefile-header to generate makefile stubs],
  [
    if test "x$withval" = "xno" ; then
      USE_GLOBUS_GPT=.pre
      GLOBUS_MAKEFILE_HEADER=:
    else
      AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/bin/globus-makefile-header, not found, $GLOBUS_LOCATION/bin)
      if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/sbin/globus-makefile-header, not found, $GLOBUS_LOCATION/sbin)
	if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	  USE_GLOBUS_GPT=.pre
	  GLOBUS_MAKEFILE_HEADER=:
	else
	  USE_GLOBUS_GPT=
	fi
      else
	USE_GLOBUS_GPT=
      fi
    fi
  ],
  [
    AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/bin/globus-makefile-header, not found, $GLOBUS_LOCATION/bin)
    if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
      AC_CHECK_PROG(GLOBUS_MAKEFILE_HEADER, globus-makefile-header, $GLOBUS_LOCATION/sbin/globus-makefile-header, not found, $GLOBUS_LOCATION/sbin)
      if test "x$GLOBUS_MAKEFILE_HEADER" = "xnot found"; then
	USE_GLOBUS_GPT=.pre
	GLOBUS_MAKEFILE_HEADER=:
      else
	USE_GLOBUS_GPT=
      fi
    else
      USE_GLOBUS_GPT=
    fi
  ]
)
AC_SUBST(GLOBUS_MAKEFILE_HEADER)

AC_ARG_WITH(flavor,
  [  --with-flavor=<flavor>               Specify the globus build flavor],
  [
    GLOBUS_FLAVOR_NAME=$withval
    if test ! -f "$GLOBUS_LOCATION/etc/globus_core/flavor_$GLOBUS_FLAVOR_NAME.gpt"; then
      echo "ERROR: Flavor $GLOBUS_FLAVOR_NAME has not been installed" >&2
      exit 1
    fi
  ],
  [
    if test "x$GLOBUS_FLAVOR_NAME" = "x"; then
      AC_CHECK_PROG(GLOBUS_FLAVOR_CONF, globus-flavor-configuration, $GLOBUS_LOCATION/sbin/globus-flavor-configuration, not found, $GLOBUS_LOCATION/sbin)
      if test "x$GLOBUS_FLAVOR_CONF" = "xnot found"; then
	AC_CHECK_PROG(GPT_FLAVOR_CONF, gpt-flavor-configuration, $GPT_LOCATION/sbin/gpt-flavor-configuration, not found, $GPT_LOCATION/sbin)
	if test "x$GPT_FLAVOR_CONF" = "xnot found"; then
	  echo "Please specify a globus build flavor" >&2
  	  exit 1
	fi
	FLAVOR_CONF=$GPT_FLAVOR_CONF
      else
	FLAVOR_CONF=$GLOBUS_FLAVOR_CONF
      fi
      GLOBUS_FLAVOR_NAME=`$FLAVOR_CONF | grep -v '^ ' | cut -f1 -d: | grep thr | tail -1`
    fi
  ]
)
AC_MSG_CHECKING(for Globus flavor)
AC_MSG_RESULT($GLOBUS_FLAVOR_NAME)
AC_SUBST(GLOBUS_FLAVOR_NAME)

GLOBUS_INCLUDES=-I$GLOBUS_LOCATION/include/$GLOBUS_FLAVOR_NAME
AC_SUBST(GLOBUS_INCLUDES)

CPPFLAGS=$GLOBUS_INCLUDES
AC_CHECK_HEADERS([globus_gsi_proxy.h])

AC_MSG_CHECKING(for globus_io version)
IO_VERSION=`$GPT_LOCATION/sbin/gpt-query -name globus_io 2>/dev/null | \
  grep 'ver:' 2>/dev/null | head -n 1 2>/dev/null | \
  sed 's/^.*ver: \([[^ ]]*\).*/\1/' 2>/dev/null`
IO_VERSION_MAJOR=`echo "$IO_VERSION" | sed 's/^\([^.]\).*/\1/'`;
if test "x$IO_VERSION_MAJOR" = "x"; then
  IO_VERSION_MAJOR=0
fi
if test "$IO_VERSION_MAJOR" -lt "4"; then
  AC_DEFINE(OLD_GLOBUS_IO,,[Globus IO version])
fi
AC_MSG_RESULT($IO_VERSION)

AC_ARG_ENABLE(experimental,
  [  --enable-experimental               Enable experimental code],  
  [
    case "${enableval}" in
      yes) experimental=true ;;
      no)  experimental=false ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --enable-experimental) ;;
    esac
  ],
  [ experimental=false ]
)
experimental_prefix="# "
if test "x$experimental" = "xtrue"; then
dnl  AC_MSG_NOTICE([Turning on experimental code])
  experimental_prefix=""
fi
AC_SUBST(experimental_prefix)
AM_CONDITIONAL(EXPERIMENTAL, test x$experimental = xtrue)

dnl Add $GLOBUS_LOCATION/lib to search for system dynamic libraries.
dnl Is used to avoid rpath of the specified paths
  
AC_ARG_ENABLE(rpath-hack,
  [  --enable-rpath-hack                 Do not rpath Globus libraries in binaries],
  [
    case "${enableval}" in
      yes) rpath_hack=true ;;
      no)  rpath_hack=false ;;
      *)   AC_MSG_ERROR(bad value ${enableval} for --enable-rpath-hack) ;;
    esac
  ],
  [ rpath_hack=true ]
)
if test "x$rpath_hack" = "xtrue"; then
dnl  AC_MSG_NOTICE([Adding GLOBUS_LOCATION/lib to system dynamic library search path])
  cp -p libtool libtool.orig
  cat libtool.orig | \
    sed "s,\(^sys_lib_dlsearch_path_spec=.*\).$,\1 $GLOBUS_LOCATION/lib\"," > \
    libtool
fi

]
)
