#include "../std.h"
#include <iostream>

#include "../misc/condition.h"
#include "../misc/log_time.h"
#include "../https/se/ns.nsmap"
#include "../https/client/client.h"
#include "databufferpar.h"
#include "datapoint.h"
#include "datahandle.h"

ns__fileinfo::ns__fileinfo() {
}
ns__fileinfo::~ns__fileinfo() {
}

class httpg_info_t;
static int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg);
static void* read_thread(void *arg);
static void* write_thread(void *arg);

typedef struct {
  httpg_info_t* stat;
  unsigned long long offset;
//  unsigned long long end;
  char* buffer;
  int buffer_h;
  unsigned int buffer_size;
  unsigned int buffer_used;
  pthread_t thr;
  HTTP_Client* s;
} httpg_state_t;

class httpg_info_t {
 friend int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg);
 friend void* read_thread(void *arg);
 friend void* write_thread(void *arg);
 friend class DataHandle;
 private:
  int streams;
  unsigned int chunk_size;
  unsigned long long int requested_size;
  int threads;
  bool have_threads;
  CondSimple lock;
  DataBufferPar* buffer;
  const char* url;
  httpg_state_t* channels;
  bool cancel;
  unsigned long long int content_size;
  bool have_content_size;
 public:
  httpg_info_t(void);
  ~httpg_info_t(void);
};

httpg_info_t::httpg_info_t(void) {
  url=NULL;
  buffer=NULL;
  threads=0; have_threads=false;
  channels=NULL;
  cancel=false;
  have_content_size=false;
  content_size=0;
}

httpg_info_t::~httpg_info_t(void) {
  if(channels) { free(channels); channels=NULL; };
}


static int read_callback(unsigned long long offset,unsigned long long size,char* buf,void* arg) {
cerr<<"read_callback: offset: "<<offset<<endl;
cerr<<"read_callback: size: "<<size<<endl;
  httpg_state_t* tstat = (httpg_state_t*)arg;
  if(tstat->buffer == NULL) return -1;
  if(tstat->buffer_used == 0) {
    tstat->offset=offset;
  };
  // move data from buf to buffer
  for(;;) {
    if(size==0) break;
    unsigned long long l = (tstat->buffer_size - tstat->buffer_used);
    if(l>size) l=size; 
cerr<<"read_callback: moving "<<l<<" bytes"<<endl;
    memcpy(tstat->buffer+tstat->buffer_used,buf,l);
    tstat->buffer_used+=size;
    size-=l; offset+=l; 
cerr<<"read_callback: buffer_used "<<tstat->buffer_used<<endl;
cerr<<"read_callback: offset: "<<offset<<endl;
cerr<<"read_callback: size: "<<size<<endl;
    if(tstat->buffer_used == tstat->buffer_size) { // full buffer
cerr<<"read_callback: full buffer"<<endl;
      if(!tstat->stat->buffer->is_read(tstat->buffer_h,tstat->buffer_used,tstat->offset)) {
cerr<<"read_callback: is_read failed"<<endl;
        tstat->buffer=NULL; tstat->buffer_h=-1;
        tstat->stat->buffer->error_read(true);
        return -1;

      };
      tstat->offset+=tstat->buffer_used;
cerr<<"read_callback: offset(tstat): "<<tstat->offset<<endl;
      // new buffer needed
      if(!tstat->stat->buffer->for_read(tstat->buffer_h,tstat->buffer_size,true)) {
cerr<<"read_callback: for_read failed"<<endl;
        tstat->buffer=NULL; tstat->buffer_h=-1;
        return -1;
      };
      tstat->buffer=(*(tstat->stat->buffer))[tstat->buffer_h];
      tstat->buffer_used=0;
    };
  };
  return 0;
}

static void* read_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
cerr<<"read_thread: threads: "<<istat->threads<<endl;
  HTTP_Client s(istat->url);
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;
  if(s) for(;;) {
    // get first buffer
cerr<<"read_thread: calling for_read"<<endl;
    if(!istat->buffer->for_read(tstat->buffer_h,tstat->buffer_size,true)) break;
cerr<<"read_thread: for_read: buffer: "<<tstat->buffer_h<<endl;
cerr<<"read_thread: for_read: size: "<<tstat->buffer_size<<endl;
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    tstat->buffer_used=0;
    // connect
cerr<<"read_thread: calling connect"<<endl;
    if(s.connect() != 0) break;
    // acquire chunk
    istat->lock.block();
    tstat->offset = istat->requested_size;
    // check for eof and cancel
    if((istat->cancel) || 
       (istat->have_content_size && (tstat->offset >= istat->content_size))) {
cerr<<"read_thread: exiting due to eof or cancel: "<<tstat->offset<<" - "<<istat->content_size<<endl;
      istat->lock.unblock();
      istat->buffer->is_read(tstat->buffer_h,0,tstat->offset);
      if(!istat->cancel) failed=false;
      break;
    };
    istat->requested_size+=istat->chunk_size;
    istat->lock.unblock();
    unsigned long long int offset = tstat->offset;
    // recieve chunk
cerr<<"read_thread: calling GET: offset: "<<tstat->offset<<endl;
cerr<<"read_thread: calling GET: size: "<<istat->chunk_size<<endl;
    int get_res = s.GET("",tstat->offset,istat->chunk_size,&read_callback,tstat);
    // accept last received data
    if(tstat->buffer) {
cerr<<"read_thread: calling is_read: "<<tstat->buffer_used<<endl;
      istat->buffer->is_read(tstat->buffer_h,tstat->buffer_used,tstat->offset);
    };
    tstat->offset+=tstat->buffer_used;

    if(get_res != 0) { // failed
cerr<<"read_thread: GET failed"<<endl;
      istat->buffer->error_read(true);
      break;
    } else {
      unsigned long long int cs = s.size();
      if(cs) {
        istat->content_size=cs; istat->have_content_size=true;
      };
      // if 0 bytes transfered - behind eof
cerr<<"read_thread: check for eof: "<<offset<<" - "<<tstat->offset<<endl;
      if(offset == tstat->offset) { failed=false; break; }; // eof
    };
  };
cerr<<"read_thread: loop exited"<<endl;
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
cerr<<"read_thread: last thread: failure: "<<failed<<endl;
    if(failed) istat->buffer->error_read(true);
    istat->buffer->eof_read(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_reading_httpg(DataBufferPar &buf) {
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=c_url.c_str();
cerr<<"start_reading_httpg: url: "<<httpg_stat->url<<endl;
  httpg_stat->chunk_size=buf.buffer_size();
cerr<<"start_reading_httpg: chunk_size: "<<httpg_stat->chunk_size<<endl;
  httpg_stat->streams=ftp_threads;
cerr<<"start_reading_httpg: streams: "<<httpg_stat->streams<<endl;
  httpg_stat->requested_size=0; // amount of data currently beeing downloaded
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
cerr<<"start_reading_httpg: channels: "<<(int)httpg_stat->channels<<endl;
  if(httpg_stat->channels==NULL) return false;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
cerr<<"start_reading_httpg: started: "<<started<<endl;
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&read_thread,httpg_stat) != 0) break;
  };
  pthread_attr_destroy(&attr);
cerr<<"start_reading_httpg: started: "<<started<<endl;
  if(started==0) return false;
  return true;
}

bool DataHandle::stop_reading_httpg(void) {
  httpg_stat->lock.block();
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  delete httpg_stat;
  return true;
}

static void* write_thread(void *arg) {
  httpg_info_t* istat = (httpg_info_t*)arg;
  istat->lock.block();
  if(istat->cancel) { istat->lock.unblock(); return NULL; };
  httpg_state_t* tstat = istat->channels+istat->threads;
  istat->threads++; istat->have_threads=true;
cerr<<"write_thread: threads: "<<istat->threads<<endl;
  HTTP_Client s(istat->url);
  tstat->s=&s;
  istat->lock.unblock();
  tstat->stat=istat;
  bool failed = true;

  if(s) for(;;) {
    // get first buffer
cerr<<"write_thread: calling for_write"<<endl;
    if(!istat->buffer->for_write(tstat->buffer_h,tstat->buffer_size,tstat->offset,true)) {
      if(!istat->buffer->error()) failed=false; // eof
      break;
    };
cerr<<"write_thread: for_write: buffer: "<<tstat->buffer_h<<endl;
cerr<<"write_thread: for_write: size: "<<tstat->buffer_size<<endl;
cerr<<"write_thread: for_write: offset: "<<tstat->offset<<endl;
    tstat->buffer=(*(istat->buffer))[tstat->buffer_h];
    // connect
cerr<<"write_thread: calling connect"<<endl;
    if(s.connect() != 0) break;
    // send chunk
    int get_res = s.PUT("",tstat->offset,tstat->buffer_size,
                             (unsigned char*)(tstat->buffer),
                             istat->have_content_size?istat->content_size:0);
    if(get_res != 0) { // failed
cerr<<"write_thread: PUT failed"<<endl;
      istat->buffer->is_notwritten(tstat->buffer_h);
      istat->buffer->error_write(true);
      break;
    } else {
cerr<<"write_thread: PUT succeeded"<<endl;
      istat->buffer->is_written(tstat->buffer_h);
    };
  };
cerr<<"write_thread: loop exited"<<endl;
  istat->lock.block();
  istat->threads--;
  if(istat->threads==0) {
cerr<<"write_thread: last thread: failure: "<<failed<<endl;
    if(failed) istat->buffer->error_write(true);
    istat->buffer->eof_write(true);
  };
  tstat->s=NULL;
  istat->lock.signal_nonblock();
  istat->lock.unblock();
  return NULL;  
}

bool DataHandle::start_writing_httpg(DataBufferPar &buf) {
  string store_url(c_url.c_str());
  // registration at SE in case of meta url ???
  if(url->meta()) {
    struct soap soap;
    HTTP_ClientSOAP s(c_url.c_str(),&soap);
    if(s.connect() != 0) {
      cerr<<"Failed to connect to "<<c_url<<endl;
      return false;
    };
    ns__fileinfo info;
    unsigned long long int size_ = url->meta_size();
    if(url->meta_size_available()) info.size=&size_;
    char md5sum_[17];
    snprintf(md5sum_,17,"%16LX",url->meta_checksum()); md5sum_[16]=0;
//    if(url->meta_checksum_available()) info.md5sum=md5sum_;
    info.md5sum=md5sum_;
    info.acl="";
    info.id=strdup(url->lfn()); 
cerr<<"start_writing_httpg: info.id: "<<info.id<<endl;
cerr<<"start_writing_httpg: lfn: "<<url->lfn()<<endl;
    ns__addResponse rr;
    int soap_err = SOAP_OK;
    string soap_url = c_url;
    int n = soap_url.find(':');
    // make gSOAP create http headers (should be removed in a future)
    if(n != string::npos) soap_url.replace(0,n,"http");
    if((soap_err=soap_call_ns__add(&soap,soap_url.c_str(),"add",&info,rr))
                                                             != SOAP_OK) {
      odlog(1)<<"Failed to execute remote soap call at "<<c_url<<endl;
      free(info.id);
      return false;
    };
    if(rr.error_code != 0) {
      odlog(1)<<"Failed ("<<rr.error_code<<") to create remote file "<<info.id<<" at "<<c_url<<endl;
      free(info.id);
      return false;
    };
    free(info.id);
    if(rr.file.url) { store_url=rr.file.url; }
    else {
      if(store_url[store_url.length()-1] != '/') store_url+="/";
      if(url->lfn()[0] == '/') {
        store_url+=(url->lfn()+1);
      } else {
        store_url+=url->lfn();
      };
    };
    odlog(2)<<"Created slot at: "<<store_url<<endl;
  };
  if(httpg_stat == NULL) httpg_stat=new httpg_info_t;
  httpg_stat->buffer=&buf; buffer=&buf;
  httpg_stat->url=strdup(store_url.c_str());
cerr<<"start_writing_httpg: url: "<<httpg_stat->url<<endl;
  httpg_stat->chunk_size=buf.buffer_size(); // not actually needed for writing
  httpg_stat->streams=ftp_threads;
cerr<<"start_writing_httpg: streams: "<<httpg_stat->streams<<endl;
  httpg_stat->requested_size=0;
  httpg_stat->threads=0;
  httpg_stat->cancel=false;
  httpg_stat->have_threads=false;
  httpg_stat->have_content_size=url->meta_size_available();
  httpg_stat->content_size=url->meta_size();
  if(httpg_stat->channels) free(httpg_stat->channels);
  httpg_stat->channels=(httpg_state_t*)malloc(sizeof(httpg_state_t)*httpg_stat->streams);
cerr<<"start_writing_httpg: channels: "<<(int)httpg_stat->channels<<endl;
  if(httpg_stat->channels==NULL) return false;
  // start threads: one per stream
  httpg_stat->lock.reset();
  int started = 0;
  pthread_attr_t attr;
  if(pthread_attr_init(&attr) != 0) return false;
  if(pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED) != 0) return false;
  for(;started < httpg_stat->streams;started++) { 
cerr<<"start_reading_httpg: started: "<<started<<endl;
    httpg_state_t* tstat = httpg_stat->channels+started;
    if(pthread_create(&(tstat->thr),&attr,&write_thread,httpg_stat) != 0) break;
  };
  pthread_attr_destroy(&attr);
cerr<<"start_writing_httpg: started: "<<started<<endl;
  if(started==0) return false;
  return true;
}

bool DataHandle::stop_writing_httpg(void) {
  httpg_stat->lock.block();
  if(!buffer->eof_read()) {
    buffer->error_read(true);
    // cancel transfer
    httpg_stat->cancel=true;
    for(int i = 0;i<httpg_stat->streams;i++) {
      if(httpg_stat->channels[i].s) httpg_stat->channels[i].s->disconnect();
    };
  };
  while(httpg_stat->threads > 0) {
    httpg_stat->lock.wait_nonblock();
  };
  httpg_stat->lock.unblock();
  free((char*)(httpg_stat->url));
  delete httpg_stat;
  return true;
}

