#ifndef __DATAMOVEPAR_H__
#define __DATAMOVEPAR_H__

#include "../std.h"
#include "../misc/condition.h"
#include "../transfer/url_map.h"

#include "datamove.h"

/*
  Wrapper around DataMove to handle few transfers at once.
*/

class DataMovePar : public DataMove {
 private:
  unsigned long long int min_speed;
  time_t min_speed_time;
  unsigned long long int min_average_speed;
  time_t max_inactivity_time;
  class DataPointPair {
   public:
    DataPointPair* next;
    DataPointPair* prev;
    DataPoint source;
    DataPoint destination;
    result res;
    bool transfer;
    bool transfered;
    bool no_cache;
    DataPointPair(const char* source_url,const char* destination_url):
                source(source_url),destination(destination_url),
                res(undefined_error),next(NULL),prev(NULL),
                transfer(false),transfered(false),no_cache(false)
    {
  };
    ~DataPointPair(void) { };
  };
  DataPointPair* points_first;
  DataPointPair* points_last;
  CondSimple transfer_cond;
  static void transfer_callback(DataMovePar* it,result res,void* arg);
 public:
  DataMovePar(void);
  ~DataMovePar(void);
  /* Add one more source and destination for list of handled transfers */
  bool Add(const char* source_url,const char* destination_url);
  /* Get source and destination from list with result of transfer */
  bool Get(string &source_url,string &destination_url,result &res);
  /* 
    Start transfer
      cache - to control data cacching
      map - to change/map source urls
      num - amount of simultaneous transfers
  */
  bool Transfer(int num = 5);
  bool Transfer(DataCache cache,const UrlMap &map,int num = 5);
};

#endif

