#include "../../std.h"

#include <ldap.h>
#include <sasl.h>

#include "se_ldap.h"

#include "se_ns_rc.h"


SENameServerRC::SENameServerRC(const char* contact,const char* se_url):url(contact),SENameServer(contact,se_url),valid(false) {
cerr<<"SENameServerRC::SENameServerRC: contact: "<<contact<<endl;
cerr<<"SENameServerRC::SENameServerRC: se_url: "<<se_url<<endl;
  if(!url) return;
  if(strcasecmp(url.Proto().c_str(),"rc")) return;
cerr<<"SENameServerRC::SENameServerRC: valid"<<endl;
  valid=true;
}

SENameServerRC::~SENameServerRC(void) {




}

int SENameServerRC::Get(SEAttributes& file) {
  int err=0;
  LDAPConnector ldap(url.Host().c_str(),url.Port());
  if(!ldap) {
    cerr<<"LDAP connection failed"<<endl;
    return -1;
  };
  string filter = "&(filename="; filter+=file.id(); filter+=")";
  err=ldap.CheckEntry(url.Path().c_str(),filter.c_str());
  if(err == -1) {
    cerr<<"LDAP query failed"<<endl;
    return -1;
  };
  if(err == 1) {
    cerr<<"LFN does not exist in RC"<<endl;
    return -1;
  };
  list<LDAPConnector::Attribute> attrs;
  attrs.push_back(LDAPConnector::Attribute("filechecksum"));
  attrs.push_back(LDAPConnector::Attribute("size"));
  attrs.push_back(LDAPConnector::Attribute("modifytime"));
  err=ldap.GetAttributes((string("rf=")+file.id()+",fin=LFROOT,"+url.Path()).c_str(),attrs);
  if(err == -1) {
    cerr<<"LDAP query failed"<<endl;
    return -1;
  };
  if(err == 1) {
    cerr<<"LFN metadata does not exist in RC"<<endl;
    return -1;
  };
  uint64_t size;
  struct tm; 
  uint8_t cksum[16];












  return err;
}

int SENameServerRC::Register(const SEFile& file,bool require_unique) {
cerr<<"SENameServerRC::Register: "<<file.id()<<endl;
  int err=0;
  LDAPConnector ldap(url.Host().c_str(),url.Port());
  if(!ldap) {
    cerr<<"LDAP connection failed"<<endl;
    return -1;
  };
  // store LFN
  list<LDAPConnector::Attribute> attrs;
  attrs.push_back(LDAPConnector::Attribute("filename",file.id()));
cerr<<"SENameServerRC::Register: SetAttributes: base: "<<url.Path()<<endl;
  err=ldap.SetAttributes(url.Path().c_str(),attrs);
cerr<<"SENameServerRC::Register: SetAttributes: err: "<<err<<endl;
  if(err==-1) return -1;
  if(err == 1) {
cerr<<"SENameServerRC::Register: LFN already present"<<endl;
    // if LFN present compare metadata
    if(require_unique) return -1;
    string filter = "&";
    char buf[100];
    snprintf(buf,99,"%Lu",file.size());
    filter+="(size="; filter+=buf; filter+=")";
    if(file.checksum_available()) {
      const char* s = file.checksum().c_str();
      if(strncmp(s,"cksum:",6) == 0) {
        filter+="(filechecksum="; filter+=(s+6); filter+=")";
      };
    };
    snprintf(buf,99,"%u",timegm((struct tm*)file.created()));
    filter+="(modifytime="; filter+=buf; filter+=")";
    err=ldap.CheckEntry((string("rf=")+file.id()+",fin=LFROOT,"+url.Path()).c_str(),filter.c_str());
    if(err != 0) return -1;
  } else {
    // store metadata
cerr<<"SENameServerRC::Register: Storing metadata"<<endl;
    list<LDAPConnector::Attribute> attrs;
    attrs.push_back(LDAPConnector::Attribute("objectClass","top"));
    attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusTop"));
    attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusReplicaFileGroup"));
    err=ldap.CreateEntry(("fin=LFROOT,"+url.Path()).c_str(),attrs);
cerr<<"SENameServerRC::Register: CreateEntry: err: "<<err<<endl;
    if(err == -1) return -1; 
    attrs.clear();
    attrs.push_back(LDAPConnector::Attribute("objectClass","top"));
    attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusTop"));
    attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusReplicaLogicalFile"));
    char buf[100];
//    for(int i = 0;i<16;i++) snprintf(buf+(i*2),99-(i*2),"%02x",file.md5()[i]);
//    attrs.push_back(LDAPConnector::Attribute("filechecksum",buf));
    snprintf(buf,99,"%Lu",file.size());
    attrs.push_back(LDAPConnector::Attribute("size",buf));
//    snprintf(buf,99,"%u",timegm((struct tm*)file.created()));
//    attrs.push_back(LDAPConnector::Attribute("modifytime",buf));
    err=ldap.CreateEntry((string("rf=")+file.id()+",fin=LFROOT,"+url.Path()).c_str(),attrs);
    if(err == -1) return -1; 
    if(err == 1) {  // inconsistency - existing metadata for non-existent LFN
      return -1;
    };
  };
  // create location
cerr<<"SENameServerRC::Register: se_url: "<<se_url()<<endl;
  string se_host = se_url();
  int n = se_host.find(':');
  if(n != string::npos) se_host.erase(0,n+3);
  n = se_host.find('/');
  if(n != string::npos) se_host.erase(n,string::npos);
cerr<<"SENameServerRC::Register: se_host: "<<se_host<<endl;
  attrs.clear();
  attrs.push_back(LDAPConnector::Attribute("objectClass","top"));
  attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusTop"));
  attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusReplicaInfo"));
  attrs.push_back(LDAPConnector::Attribute("objectClass","GlobusReplicaLocation"));
  attrs.push_back(LDAPConnector::Attribute("uc",se_url()));
  attrs.push_back(LDAPConnector::Attribute("filename",file.id()));
  err=ldap.CreateEntry((string("re=")+se_host+","+url.Path()).c_str(),attrs);
  if(err == -1) return -1;
  if(err == 1) {
    // store location
    attrs.clear();
    attrs.push_back(LDAPConnector::Attribute("filename",file.id()));
    err=ldap.SetAttributes((string("re=")+se_host+","+url.Path()).c_str(),attrs);
    if(err==-1) return -1;
  };
  return 0;
}

