#include "../../std.h"
#include <arpa/inet.h>

#include <globus_common.h>

#include "../../misc/log_time.h"
#include "../../misc/inttostring.h"
#include "../../config/environment.h"

#include "../httpsd.h"

static int curr_connections = 0;
static int pre_connections = 2;
static int max_connections = 100;

static char* service_not_found_str = "Service not found\r\n";

list<HTTPS_Connector*> HTTPS_Connector::cons;
pthread_mutex_t HTTPS_Connector::cons_lock = PTHREAD_MUTEX_INITIALIZER;

extern string base_ssl_url;
extern string base_gsi_url;

void HTTPS_Connector::loop(void) {
  char buf[1024];
  {
    int host[4];
    unsigned short port;
    if(globus_io_tcp_get_local_address(&h,host,&port) == GLOBUS_SUCCESS) {
cerr<<"host: "<<host[0]<<"."<<host[1]<<"."<<host[2]<<"."<<host[3]<<":"<<port<<endl;
//    if(getsockname(h.fd,  struct  sockaddr  *  name  , socklen_t * namelen )
      struct hostent he;
      struct hostent* he_p;
      char abuf[1024]; abuf[sizeof(abuf)-1]=0;
      snprintf(abuf,sizeof(abuf)-1,"%u.%u.%u.%u",
                              host[0],host[1],host[2],host[3]);
cerr<<"host printed: "<<abuf<<endl;
      struct in_addr a;
      if(inet_aton(abuf,&a) != 0) {
        int h_errnop;
        he_p=globus_libc_gethostbyaddr_r((char*)&a,strlen(abuf),AF_INET,&he,buf,sizeof(buf),&h_errnop);
        if(he_p) {
cerr<<"host resolved: "<<he_p->h_name<<endl;
          if(strcmp(he_p->h_name,"localhost") == 0) {
            abuf[sizeof(abuf)-1]=0;
            if(globus_libc_gethostname(abuf,sizeof(abuf)-1) != 0) {
              strcpy(abuf,"localhost");
            };
//            default_base_url=string("httpg://")+abuf+":"+inttostring(port);
          } else {
//            default_base_url=string("httpg://")+he_p->h_name+":"+inttostring(port);
          };
        };
      };
    };
//    if(default_base_url.length() == 0) {
//      default_base_url="httpg://localhost:"+inttostring(port);
//    };
  };
  int keep_alive = 1;
  for(;;) {
    /* read first line of request header */
cerr<<"---- read first line of request header ----"<<endl;
    size_t l = readline(buf,sizeof(buf)-1);
    if(l == 0) break;
cerr<<"---- first line: "<<buf<<endl;
    // parse request
    char* http_URI;
    char* http_version;
    char* http_method = buf;
    char* t = buf;
    t=strchr(t,' ');
    if(t == NULL) {
      olog<<"Broken HTTP request (no URI found) - "<<buf<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    http_URI=t;
    t=strchr(t,' ');
    if(t == NULL) {
      olog<<"Broken HTTP request (no version found) - "<<http_URI<<endl; break;
    };
    (*t)=0; t++; for(;(*t)==' ';t++);
    if(strncmp("HTTP/",t,5) != 0) {
      olog<<"Unknown protocol: "<<t<<endl; break;
    };
    http_version=t+5;
    if(keep_alive) {
      if(strcmp(http_version,"1.0") == 0) {
        keep_alive=1;
      } else {
        keep_alive=2;
      };
    };
    HTTP_Error err = HTTP_OK;
    // adjust URI
    string http_url = http_URI;
    if(strncasecmp("https://",http_url.c_str(),8) &&
       strncasecmp("httpg://",http_url.c_str(),8)) {
      if(http_url[0] != '/') http_url="/"+http_url;
      http_url=base_url+http_url;
    };
    if(http_url[http_url.length()-1] == '/') http_url.resize(http_url.length()-1);
    olog<<"Request on "<<http_url<<endl;
    // make sure request is for proper url
    if(strncasecmp(http_url.c_str(),base_url.c_str(),base_url.length()) != 0) {
      olog<<"URL does not correspond to conenction: "<<base_url<<endl;
      err=HTTP_NOT_FOUND;
    } else {
      // obtain service to use 
      HTTP_Service* service = services.get(*this,http_url.c_str());
      if(service == NULL) {
        olog<<"Service not found: "<<http_url<<endl;
        err=HTTP_NOT_FOUND;
      } else {
        if(strcasecmp(http_method,"GET") == 0) {
          cerr<<"Calling GET"<<endl;
          err=service->get(http_url.c_str(),keep_alive);
        } else if(strcasecmp(http_method,"PUT") == 0) {
          cerr<<"Calling PUT"<<endl;
          err=service->put(http_url.c_str(),keep_alive);
        } else if(strcasecmp(http_method,"POST") == 0) {
          cerr<<"Calling POST"<<endl;
          err=service->post(http_url.c_str(),keep_alive);
        } else {
          cerr<<"Unknown method: "<<http_method<<endl;
          err=HTTP_NOT_IMPLEMENTED;
        };
      };
    };
    if((err==HTTP_NOT_IMPLEMENTED) || (err==HTTP_NOT_FOUND)) { 
      if(skip_request(*this,keep_alive) != 0) break;
    };
    cerr<<"---- keep_alive: "<<keep_alive<<endl;
    if((err != HTTP_OK) && (err != HTTP_FAILURE)) { /* send error response */
      cerr<<"---- send_error_response ----"<<endl;
      if(send_error_response(*this,keep_alive,err,NULL,NULL) != 0) {
        cerr<<"Failed sending response"<<endl; break;
      };
    };
    cerr<<"---- checking keep_alive: "<<keep_alive<<endl;
    if(keep_alive == 0) break;
    cerr<<"---- looping ----"<<endl;
  };
  olog<<"Exiting\n";
  // close connection
  services.close(*this);
}

void HTTPS_Connector::accept_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result) {
  if(result != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(result));
    cerr<<"Failed to accept connection: "<<tmp<<endl;
    free(tmp);
    return;
  };
  HTTPS_Connector* it = (HTTPS_Connector*)callback_arg;
  globus_thread_blocking_will_block();  // instead of creating new thread
  it->loop();
  delete it;
}

HTTPS_Connector::HTTPS_Connector(globus_io_handle_t *s,const char* url,HTTP_Services& serv,list<AuthEvaluator*>& auths_):services(serv),auths(auths_) {
  if(url) base_url=url;
  pthread_mutex_lock(&cons_lock);
  cons.push_back(this);
  pthread_mutex_unlock(&cons_lock);
  initialized=false;
  globus_result_t res;

  res=globus_io_tcp_register_accept(s,GLOBUS_NULL,&h,&accept_cb,this);
  if(res != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(res));
    cerr<<"Failed to accept connection: "<<tmp<<endl; free(tmp);
    return;
  };
  initialized=true;
  return;
}

HTTPS_Connector::~HTTPS_Connector(void) {
  pthread_mutex_lock(&cons_lock);
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if((*i)==this) { cons.erase(i); break; };
  };
  pthread_mutex_unlock(&cons_lock);
  if(initialized) globus_io_close(&h);
}

void HTTPS_Connector::identity(globus_io_handle_t* handle,const char* subject,gss_cred_id_t cred) {
  pthread_mutex_lock(&cons_lock);
  for(list<HTTPS_Connector*>::iterator i = cons.begin();i!=cons.end();++i) {
    if(&((*i)->h) == handle) {
      (*i)->user.set(subject,cred);
      cerr<<"Subject is "<<(*i)->user.DN()<<endl;
      cerr<<"Proxy stored to "<<(*i)->user.proxy()<<endl;
      break;
    };
  };
  pthread_mutex_unlock(&cons_lock);
}

size_t HTTPS_Connector::readline(char* buf, size_t l) {
// cerr<<"readline: "<<l<<endl;
  char c;
  unsigned int ll = 0;
  for(;;) {
    if(read(&c,1) != 1) return 0;
    if(ll<l) buf[ll]=c;
    ll++;
    if(c == 0x0a) break;
  };
  if(ll > l) ll=l;
  l=ll-1;
  for(;l>=0;l--) {
    if((buf[l] != 0x0a) && (buf[l] != 0x0d)) break;
    buf[l]=0;
  };
// cerr<<"readline: "<<buf<<endl;
  return ll;
}

size_t HTTPS_Connector::read(char* buf, size_t l) {
  globus_result_t res;
  globus_size_t ll = 0;
  if((res=globus_io_read(&h,(globus_byte_t*)buf,l,1,&ll)) != GLOBUS_SUCCESS) {
    char* tmp = globus_object_printable_to_string(globus_error_get(res));
//    fprintf(stderr,"frecv: %s\n",tmp); free(tmp);
    return 0;
  };
  for(int ii=0;ii<ll;ii++) fprintf(stdout,"%c",buf[ii]);
  return ll;
}

int HTTPS_Connector::write(const char* buf, size_t l) {
  size_t n = 0;
//  for(int ii=0;ii<l;ii++) fprintf(stdout,"%c",buf[ii]);
  for(;n<l;) {
    globus_size_t ll = 0;
    globus_result_t res;
    if((res=globus_io_write(&h,(globus_byte_t*)buf+n,l-n,&ll)) != GLOBUS_SUCCESS) {
      char* tmp = globus_object_printable_to_string(globus_error_get(res));
      fprintf(stderr,"fsend: %s\n",tmp); free(tmp);
      return -1;
    };
    n+=ll;
  };
  return 0;
}

int skip_header(HTTPS_Connector &c,int &keep_alive) {
/* TODO - control keep_alive */
  char buf[1024], *s;
  for (;;) {
    if(c.readline(buf,1023) == 0) return 1; /* connection broken */
//    buf[1023]=0; cerr<<"skip: "<<buf<<endl;
    if(buf[0] == 0) break; /* empty line */
  };
  return 0;
}

int skip_request(HTTPS_Connector &c,int &keep_alive) {
  bool length_passed=false;
  unsigned long long int length;
  char buf[1024];
  // skip header
  for (;;) {
    if(c.readline(buf,1023) == 0) return HTTP_ERROR; /* connection broken */
    buf[1023]=0; cerr<<"header: "<<buf<<endl;
    if(buf[0] == 0) break; /* empty line - end of header */
    char* token = buf; for(;*token;token++) if(isspace(*token)) break;
    int l = token-buf;
    for(;*token;token++) if(!isspace(*token)) break;
    if(strncasecmp("Connection:",buf,l) == 0) {
      if(strcasecmp("close",token) == 0) { keep_alive=0; return HTTP_OK; }
      else if(strcasecmp("keep-alive",token) == 0) { keep_alive=2; };
    } else if(strncasecmp("Content-Length:",buf,l) == 0) {
      char *e;
      length=strtoull(token,&e,10);
      if((*e) == 0) length_passed=true;
      if(length > 1024) { keep_alive=0; return HTTP_OK; };        
    };
  };
  // skip body
  if(!length_passed) { keep_alive=0; return HTTP_OK; };
  for (;;) {
    size_t l = 1024;
    if(l>length) l=length;
    l=c.read(buf,0);
    if(l==0) { keep_alive=0; return HTTP_OK; };
    length-=l; 
    if(length == 0) break;
  };
  return HTTP_OK;
}

int send_error_response(HTTPS_Connector &c,int keep_alive,int code,char* type,char* content) {
  unsigned long long int size = 0;
  if(content) {
    size=strlen(content);
    if(send_response_header(c,keep_alive,code,type,size) != 0) return -1;
    return c.write(content,strlen(content));
  } else {
    string fname = nordugrid_loc+"/share/error"+inttostring(code)+".html";
    stat_file(fname.c_str(),size);
    if(send_response_header(c,keep_alive,code,"text/html",size) != 0) return -1;
    if(size) return send_file(fname.c_str(),c);
  };
  return 0;
}

int send_response_header(HTTPS_Connector &c,int keep_alive,int code,char* type,int size) {
  string header = "HTTP/1.1 "+inttostring(code)+" OK\r\n";
  if(keep_alive && (size>=0)) {
    header+="Connection: Keep-Alive\r\n";
  } else {
    header+="Connection: close\r\n";
  };
  if(size>=0) header+="Content-Length: "+inttostring(size)+"\r\n";
  if(type) { header+="Content-Type: "; header+=type; header+="\r\n"; };
  header+="\r\n";
  olog<<"Sending response:\n"<<header<<endl;
  return c.write(header.c_str(),header.length());
}


